package java.util.concurrent;

import java.util.function.Consumer;
import java.util.Spliterators;
import java.util.NoSuchElementException;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Spliterator;
import java.lang.reflect.Array;
import java.util.Iterator;
import java.util.Collection;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.atomic.AtomicInteger;
import java.io.Serializable;
import java.util.AbstractQueue;

public class LinkedBlockingQueue<E> extends AbstractQueue<E> implements BlockingQueue<E>, Serializable
{
    private static final long serialVersionUID = -6903933977591709194L;
    private final int capacity;
    private final AtomicInteger count;
    transient LinkedBlockingQueue.Node<E> head;
    private transient LinkedBlockingQueue.Node<E> last;
    private final ReentrantLock takeLock;
    private final Condition notEmpty;
    private final ReentrantLock putLock;
    private final Condition notFull;
    
    private void signalNotEmpty() {
        ReentrantLock takeLock = this.takeLock;
        takeLock.lock();
        try {
            this.notEmpty.signal();
        }
        finally {
            takeLock.unlock();
        }
    }
    
    private void signalNotFull() {
        ReentrantLock putLock = this.putLock;
        putLock.lock();
        try {
            this.notFull.signal();
        }
        finally {
            putLock.unlock();
        }
    }
    
    private void enqueue(LinkedBlockingQueue.Node<E> node) {
        this.last.next = node;
        this.last = node;
    }
    
    private E dequeue() {
        LinkedBlockingQueue.Node<Object> head = this.head;
        LinkedBlockingQueue.Node<Object> next = head.next;
        head.next = head;
        this.head = next;
        Object item = next.item;
        next.item = null;
        return item;
    }
    
    void fullyLock() {
        this.putLock.lock();
        this.takeLock.lock();
    }
    
    void fullyUnlock() {
        this.takeLock.unlock();
        this.putLock.unlock();
    }
    
    public LinkedBlockingQueue() {
        this(Integer.MAX_VALUE);
    }
    
    public LinkedBlockingQueue(int capacity) {
        this.count = new AtomicInteger();
        this.takeLock = new ReentrantLock();
        this.notEmpty = this.takeLock.newCondition();
        this.putLock = new ReentrantLock();
        this.notFull = this.putLock.newCondition();
        if (capacity <= 0) {
            throw new IllegalArgumentException();
        }
        this.capacity = capacity;
        LinkedBlockingQueue.Node node = new LinkedBlockingQueue.Node(null);
        this.head = node;
        this.last = node;
        return;
    }
    
    public LinkedBlockingQueue(Collection<? extends E> collection) {
        this(Integer.MAX_VALUE);
        ReentrantLock putLock = this.putLock;
        putLock.lock();
        try {
            int newValue = 0;
            for (Object next : collection) {
                if (next == null) {
                    throw new NullPointerException();
                }
                if (newValue == this.capacity) {
                    throw new IllegalStateException("Queue full");
                }
                this.enqueue(new LinkedBlockingQueue.Node(next));
                ++newValue;
            }
            this.count.set(newValue);
        }
        finally {
            putLock.unlock();
        }
    }
    
    @Override
    public int size() {
        return this.count.get();
    }
    
    @Override
    public int remainingCapacity() {
        return this.capacity - this.count.get();
    }
    
    @Override
    public void put(E e) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        int andIncrement = -1;
        LinkedBlockingQueue.Node<E> node = new LinkedBlockingQueue.Node(e);
        ReentrantLock putLock = this.putLock;
        AtomicInteger count = this.count;
        putLock.lockInterruptibly();
        try {
            while (count.get() == this.capacity) {
                this.notFull.await();
            }
            this.enqueue(node);
            andIncrement = count.getAndIncrement();
            if (andIncrement + 1 < this.capacity) {
                this.notFull.signal();
            }
        }
        finally {
            putLock.unlock();
        }
        if (andIncrement == 0) {
            this.signalNotEmpty();
        }
        return;
    }
    
    @Override
    public boolean offer(E e, long n, TimeUnit timeUnit) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        long n2 = timeUnit.toNanos(n);
        int andIncrement = -1;
        ReentrantLock putLock = this.putLock;
        AtomicInteger count = this.count;
        putLock.lockInterruptibly();
        try {
            while (count.get() == this.capacity) {
                if (n2 <= 0L) {
                    return false;
                }
                n2 = this.notFull.awaitNanos(n2);
            }
            this.enqueue(new LinkedBlockingQueue.Node(e));
            andIncrement = count.getAndIncrement();
            if (andIncrement + 1 < this.capacity) {
                this.notFull.signal();
            }
        }
        finally {
            putLock.unlock();
        }
        if (andIncrement == 0) {
            this.signalNotEmpty();
        }
        return true;
    }
    
    @Override
    public boolean offer(E e) {
        if (e == null) {
            throw new NullPointerException();
        }
        AtomicInteger count = this.count;
        if (count.get() == this.capacity) {
            return false;
        }
        int andIncrement = -1;
        LinkedBlockingQueue.Node<E> node = new LinkedBlockingQueue.Node(e);
        ReentrantLock putLock = this.putLock;
        putLock.lock();
        try {
            if (count.get() < this.capacity) {
                this.enqueue(node);
                andIncrement = count.getAndIncrement();
                if (andIncrement + 1 < this.capacity) {
                    this.notFull.signal();
                }
            }
        }
        finally {
            putLock.unlock();
        }
        if (andIncrement == 0) {
            this.signalNotEmpty();
        }
        return andIncrement >= 0;
    }
    
    @Override
    public E take() throws InterruptedException {
        int andDecrement = -1;
        AtomicInteger count = this.count;
        ReentrantLock takeLock = this.takeLock;
        takeLock.lockInterruptibly();
        Object dequeue;
        try {
            while (count.get() == 0) {
                this.notEmpty.await();
            }
            dequeue = this.dequeue();
            andDecrement = count.getAndDecrement();
            if (andDecrement > 1) {
                this.notEmpty.signal();
            }
        }
        finally {
            takeLock.unlock();
        }
        if (andDecrement == this.capacity) {
            this.signalNotFull();
            return dequeue;
        }
        return dequeue;
    }
    
    @Override
    public E poll(long n, TimeUnit timeUnit) throws InterruptedException {
        Object dequeue = null;
        int andDecrement = -1;
        long n2 = timeUnit.toNanos(n);
        AtomicInteger count = this.count;
        ReentrantLock takeLock = this.takeLock;
        takeLock.lockInterruptibly();
        try {
            while (count.get() == 0) {
                if (n2 <= 0L) {
                    return null;
                }
                n2 = this.notEmpty.awaitNanos(n2);
            }
            dequeue = this.dequeue();
            andDecrement = count.getAndDecrement();
            if (andDecrement > 1) {
                this.notEmpty.signal();
            }
        }
        finally {
            takeLock.unlock();
        }
        if (andDecrement == this.capacity) {
            this.signalNotFull();
            return dequeue;
        }
        return dequeue;
    }
    
    @Override
    public E poll() {
        AtomicInteger count = this.count;
        if (count.get() == 0) {
            return null;
        }
        Object dequeue = null;
        int andDecrement = -1;
        ReentrantLock takeLock = this.takeLock;
        takeLock.lock();
        try {
            if (count.get() > 0) {
                dequeue = this.dequeue();
                andDecrement = count.getAndDecrement();
                if (andDecrement > 1) {
                    this.notEmpty.signal();
                }
            }
        }
        finally {
            takeLock.unlock();
        }
        if (andDecrement == this.capacity) {
            this.signalNotFull();
            return dequeue;
        }
        return dequeue;
    }
    
    @Override
    public E peek() {
        if (this.count.get() == 0) {
            return null;
        }
        ReentrantLock takeLock = this.takeLock;
        takeLock.lock();
        try {
            LinkedBlockingQueue.Node<Object> next = this.head.next;
            if (next == null) {
                return null;
            }
            return next.item;
        }
        finally {
            takeLock.unlock();
        }
    }
    
    void unlink(LinkedBlockingQueue.Node<E> node, LinkedBlockingQueue.Node<E> last) {
        node.item = null;
        last.next = node.next;
        if (this.last == node) {
            this.last = last;
        }
        if (this.count.getAndDecrement() == this.capacity) {
            this.notFull.signal();
        }
    }
    
    @Override
    public boolean remove(Object o) {
        if (o == null) {
            return false;
        }
        this.fullyLock();
        try {
            LinkedBlockingQueue.Node<E> head = this.head;
            LinkedBlockingQueue.Node<E> node = head.next;
            while (node != null) {
                if (o.equals(node.item)) {
                    this.unlink(node, head);
                    return true;
                }
                head = node;
                node = node.next;
            }
            return false;
        }
        finally {
            this.fullyUnlock();
        }
    }
    
    @Override
    public boolean contains(Object o) {
        if (o == null) {
            return false;
        }
        this.fullyLock();
        try {
            LinkedBlockingQueue.Node<Object> node = this.head.next;
            while (node != null) {
                if (o.equals(node.item)) {
                    return true;
                }
                node = node.next;
            }
            return false;
        }
        finally {
            this.fullyUnlock();
        }
    }
    
    @Override
    public Object[] toArray() {
        this.fullyLock();
        try {
            Object[] array = new Object[this.count.get()];
            int n = 0;
            for (LinkedBlockingQueue.Node node = this.head.next; node != null; node = node.next) {
                array[n++] = node.item;
            }
            return array;
        }
        finally {
            this.fullyUnlock();
        }
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        this.fullyLock();
        try {
            int value = this.count.get();
            if (array.length < value) {
                array = (Object[])(Object[])Array.newInstance(array.getClass().getComponentType(), value);
            }
            int n = 0;
            for (LinkedBlockingQueue.Node node = this.head.next; node != null; node = node.next) {
                array[n++] = node.item;
            }
            if (array.length > n) {
                array[n] = null;
            }
            return array;
        }
        finally {
            this.fullyUnlock();
        }
    }
    
    @Override
    public String toString() {
        this.fullyLock();
        try {
            LinkedBlockingQueue.Node<Object> node = this.head.next;
            if (node == null) {
                return "[]";
            }
            StringBuilder sb = new StringBuilder();
            sb.append('[');
            while (true) {
                Object item = node.item;
                sb.append((item == this) ? "(this Collection)" : item);
                node = node.next;
                if (node == null) {
                    break;
                }
                sb.append(',').append(' ');
            }
            return sb.append(']').toString();
        }
        finally {
            this.fullyUnlock();
        }
    }
    
    @Override
    public void clear() {
        this.fullyLock();
        try {
            LinkedBlockingQueue.Node<Object> next;
            for (LinkedBlockingQueue.Node head = this.head; (next = head.next) != null; head = next) {
                head.next = head;
                next.item = null;
            }
            this.head = this.last;
            if (this.count.getAndSet(0) == this.capacity) {
                this.notFull.signal();
            }
        }
        finally {
            this.fullyUnlock();
        }
    }
    
    @Override
    public int drainTo(Collection<? super E> collection) {
        return this.drainTo(collection, Integer.MAX_VALUE);
    }
    
    @Override
    public int drainTo(Collection<? super E> collection, int a) {
        if (collection == null) {
            throw new NullPointerException();
        }
        if (collection == this) {
            throw new IllegalArgumentException();
        }
        if (a <= 0) {
            return 0;
        }
        int n = 0;
        ReentrantLock takeLock = this.takeLock;
        takeLock.lock();
        try {
            int min = Math.min(a, this.count.get());
            LinkedBlockingQueue.Node<Object> head = this.head;
            int i = 0;
            try {
                while (i < min) {
                    LinkedBlockingQueue.Node<Object> next = head.next;
                    collection.add(next.item);
                    next.item = null;
                    head.next = head;
                    head = next;
                    ++i;
                }
                if (i > 0) {
                    this.head = head;
                    n = (this.count.getAndAdd(-i) == this.capacity);
                }
                return min;
            }
            finally {
                if (i > 0) {
                    this.head = head;
                    n = (this.count.getAndAdd(-i) == this.capacity);
                }
            }
        }
        finally {
            takeLock.unlock();
            if (n != 0) {
                this.signalNotFull();
            }
        }
    }
    
    @Override
    public Iterator<E> iterator() {
        return new LinkedBlockingQueue.Itr();
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return new LinkedBlockingQueue.LBQSpliterator(this);
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        this.fullyLock();
        try {
            objectOutputStream.defaultWriteObject();
            for (LinkedBlockingQueue.Node node = this.head.next; node != null; node = node.next) {
                objectOutputStream.writeObject(node.item);
            }
            objectOutputStream.writeObject(null);
        }
        finally {
            this.fullyUnlock();
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        this.count.set(0);
        LinkedBlockingQueue.Node node = new LinkedBlockingQueue.Node(null);
        this.head = node;
        this.last = node;
        while (true) {
            Object object = objectInputStream.readObject();
            if (object == null) {
                break;
            }
            this.add(object);
        }
    }
    
    static class Node<E>
    {
        E item;
        LinkedBlockingQueue.Node<E> next;
        
        Node(E item) {
            this.item = item;
        }
    }
    
    private class Itr implements Iterator<E>
    {
        private LinkedBlockingQueue.Node<E> current;
        private LinkedBlockingQueue.Node<E> lastRet;
        private E currentElement;
        
        Itr() {
            LinkedBlockingQueue.this.fullyLock();
            try {
                this.current = LinkedBlockingQueue.this.head.next;
                if (this.current != null) {
                    this.currentElement = this.current.item;
                }
            }
            finally {
                LinkedBlockingQueue.this.fullyUnlock();
            }
        }
        
        @Override
        public boolean hasNext() {
            return this.current != null;
        }
        
        private LinkedBlockingQueue.Node<E> nextNode(LinkedBlockingQueue.Node<E> node) {
            while (true) {
                LinkedBlockingQueue.Node<Object> next = node.next;
                if (next == node) {
                    return LinkedBlockingQueue.this.head.next;
                }
                if (next == null || next.item != null) {
                    return next;
                }
                node = next;
            }
        }
        
        @Override
        public E next() {
            LinkedBlockingQueue.this.fullyLock();
            try {
                if (this.current == null) {
                    throw new NoSuchElementException();
                }
                Object currentElement = this.currentElement;
                this.lastRet = this.current;
                this.current = this.nextNode(this.current);
                this.currentElement = ((this.current == null) ? null : this.current.item);
                return currentElement;
            }
            finally {
                LinkedBlockingQueue.this.fullyUnlock();
            }
        }
        
        @Override
        public void remove() {
            if (this.lastRet == null) {
                throw new IllegalStateException();
            }
            LinkedBlockingQueue.this.fullyLock();
            try {
                LinkedBlockingQueue.Node<Object> lastRet = this.lastRet;
                this.lastRet = null;
                LinkedBlockingQueue.Node<Object> head = LinkedBlockingQueue.this.head;
                LinkedBlockingQueue.Node<Object> node = head.next;
                while (node != null) {
                    if (node == lastRet) {
                        LinkedBlockingQueue.this.unlink(node, head);
                        break;
                    }
                    head = node;
                    node = node.next;
                }
            }
            finally {
                LinkedBlockingQueue.this.fullyUnlock();
            }
            return;
        }
    }
    
    static final class LBQSpliterator<E> implements Spliterator<E>
    {
        static final int MAX_BATCH = 33554432;
        final LinkedBlockingQueue<E> queue;
        LinkedBlockingQueue.Node<E> current;
        int batch;
        boolean exhausted;
        long est;
        
        LBQSpliterator(LinkedBlockingQueue<E> queue) {
            this.queue = queue;
            this.est = (long)queue.size();
        }
        
        @Override
        public long estimateSize() {
            return this.est;
        }
        
        @Override
        public Spliterator<E> trySplit() {
            LinkedBlockingQueue<Object> queue = this.queue;
            int batch = this.batch;
            int n = (batch <= 0) ? 1 : ((batch >= 33554432) ? 33554432 : (batch + 1));
            LinkedBlockingQueue.Node<Object> node;
            if (!this.exhausted && ((node = this.current) != null || (node = queue.head.next) != null) && node.next != null) {
                Object[] array = new Object[n];
                int n2 = 0;
                LinkedBlockingQueue.Node<Object> current = this.current;
                queue.fullyLock();
                try {
                    if (current != null || (current = queue.head.next) != null) {
                        do {
                            if ((array[n2] = current.item) != null) {
                                ++n2;
                            }
                            else {
                                continue;
                            }
                        } while ((current = current.next) != null && n2 < n);
                    }
                }
                finally {
                    queue.fullyUnlock();
                }
                this.current = current;
                if (current == null) {
                    this.est = 0L;
                    this.exhausted = true;
                }
                else {
                    long est = this.est - (long)n2;
                    this.est = est;
                    if (est < 0L) {
                        this.est = 0L;
                    }
                }
                if (n2 > 0) {
                    this.batch = n2;
                    return Spliterators.spliterator(array, 0, n2, 4368);
                }
            }
            return null;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            LinkedBlockingQueue<Object> queue = this.queue;
            if (!this.exhausted) {
                this.exhausted = true;
                LinkedBlockingQueue.Node<Object> node = this.current;
                do {
                    Object item = null;
                    queue.fullyLock();
                    try {
                        if (node == null) {
                            node = queue.head.next;
                        }
                        while (node != null) {
                            item = node.item;
                            node = node.next;
                            if (item != null) {
                                break;
                            }
                            continue;
                        }
                    }
                    finally {
                        queue.fullyUnlock();
                    }
                    if (item != null) {
                        consumer.accept(item);
                    }
                    else {
                        continue;
                    }
                } while (node != null);
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            LinkedBlockingQueue<Object> queue = this.queue;
            if (!this.exhausted) {
                Object item = null;
                queue.fullyLock();
                try {
                    if (this.current == null) {
                        this.current = queue.head.next;
                    }
                    while (this.current != null) {
                        item = this.current.item;
                        this.current = this.current.next;
                        if (item != null) {
                            break;
                        }
                        continue;
                    }
                }
                finally {
                    queue.fullyUnlock();
                }
                if (this.current == null) {
                    this.exhausted = true;
                }
                if (item != null) {
                    consumer.accept(item);
                    return true;
                }
            }
            return false;
        }
        
        @Override
        public int characteristics() {
            return 4368;
        }
    }
}
